FOverhaul = FOverhaul or class()
FOverhaul._delayed_functions = FOverhaul._delayed_functions or {}
FOverhaul._update_function = FOverhaul._update_function or {}

function FOverhaul:_is_already_updating(id, func)
    for _, upd in pairs(self._update_function) do
        if upd == id or upd == func then
            return true
        end
    end

    return false
end

function FOverhaul:add_simple_updator(delay, id, func)
    if self:_is_already_updating(id, func) then
        return
    end

    table.insert(self._update_function, id)

    for i, upds in pairs(self._update_function) do
        if upds == id then
            self._update_function[i] = func
        end
    end

    if delay then
        self:add_delay_to_update(delay, func)
    end

    self:init_updates(true)
end

function FOverhaul:init_updates(updator)
    self._updating = updator

    Hooks:Add("GameSetupUpdate", "update_functions_", function(t, dt)
        if self._update_function and self._updating then
            for i, func in pairs(self._update_function) do
                if self._delayed_functions[func] then
                    self._delay = self._delay or self._delayed_functions[func]
                    if self._delay <= t then
                        func()
                        self._delay = t + self._delayed_functions[func]
                    end
                else
                    func()
                end
            end
        end
    end)
end

function FOverhaul:add_delay_to_update(delay, func)
    self._delayed_functions[func] = delay
end

function FOverhaul:remove_updates(func)
    for i, upds in pairs(self._update_function) do
        if upds == func then
            table.remove(self._update_function, i)
        end
    end
end

function FOverhaul:hold_all_updates()
    self._updating = false
end

function FOverhaul:clear_updates()
    self._update_function = {}
end

function FOverhaul:_chk_incompatibility(mod_check)
    local nonomods = { "Player Stats Panel Enhancements" }
    for _, mod in pairs(nonomods) do
        if mod_check == mod and BeardLib.Utils:ModExists(mod) then
            return true
        end
    end

    return false
end